@@ -1,28 +1,37 @@
-#' @title Optimizer using the Hyperband algorithm
+#' @title Optimizer Using the Hyperband Algorithm
 #'
 #' @name mlr_optimizers_hyperband
+#' @templateVar id hyperband
 #'
 #' @description
-#' `OptimizerHyperband` class that implements hyperband optimization. Hyperband
-#' is a budget oriented-procedure, weeding out suboptimal performing
-#' points early in a sequential training process, increasing
-#' optimization efficiency as a consequence.
+#' `OptimizerHyperband` class that implements hyperband optimization (HB).
+#' HB repeatedly calls SHA ([OptimizerSuccessiveHalving]) with different
+#' numbers of starting points. A larger number of starting points corresponds to
+#' a smaller budget allocated in the base stage. Each run of SHA within HB is
+#' called a bracket. HB considers `s_max + 1` brackets with `s_max =
+#' floor(log(r_max / r_min, eta)`. The most explorative bracket `s = s_max`
+#' constructs `s_max + 1` stages and allocates the minimum budget (`r_min`) in
+#' the base stage. The minimum budget is increased in each bracket by a factor
+#' of `eta` and the number of starting points is computed so that each bracket
+#' approximately spends the same budget. Use [hyperband_schedule()] to get a
+#' preview of the bracket layout.
 #'
-#' For this, several brackets are constructed with an associated set of
-#' points for each bracket. Each bracket as several stages. Different
-#' brackets are initialized with different amounts of points and
-#' different budget sizes. To get an idea of how the bracket layout looks like
-#' for a given argument set, use [hyperband_schedule()].
+#' | `s` |     |       |     3 |     |       |     2 |     |       |     1 |     |       |     0 |
+#' | --: | --- | ----: | ----: | --- | ----: | ----: | --- | ----: | ----: | --- | ----: | ----: |
+#' | `i` |     | `n_i` | `r_i` |     | `n_i` | `r_i` |     | `n_i` | `r_i` |     | `n_i` | `r_i` |
+#' |   0 |     |     8 |     1 |     |     6 |     2 |     |     4 |     4 |     |     8 |     4 |
+#' |   1 |     |     4 |     2 |     |     3 |     4 |     |     2 |     8 |     |       |       |
+#' |   2 |     |     2 |     4 |     |     1 |     8 |     |       |       |     |       |       |
+#' |   3 |     |     1 |     8 |     |       |       |     |       |       |     |       |       |
 #'
-#' To identify the budget for evaluating hyperband, the user has to specify
-#' explicitly which parameter of the objective function influences the budget by
-#' tagging a single parameter in the [paradox::ParamSet] with `"budget"`.
+#' `s` is the bracket number, `i` is stage number, `n_i` is the number of
+#' configurations and `r_i` is the budget allocated to a single configuration.
 #'
-#' Naturally, hyperband terminates once all of its brackets are evaluated, so a
-#' [bbotk::Terminator] in the [bbotk::OptimInstanceSingleCrit] |
-#' [bbotk::OptimInstanceMultiCrit] acts as an upper bound and should be only set to a
-#' low value if one is unsure of how long hyperband will take to finish under
-#' the given settings.
+#' The budget hyperparameter must be tagged with `"budget"` in the search space.
+#' The minimum budget (`r_min`) which is allocated in the base stage of the most
+#' explorative bracket, is set by the lower bound of the budget parameter. The
+#' upper bound defines the maximum budget (`r_max`) which which is allocated to
+#' the candidates in the last stages.
 #'
 #' @templateVar id hyperband
 #' @template section_dictionary_optimizers
@@ -30,18 +39,20 @@
 #' @section Parameters:
 #' \describe{
 #' \item{`eta`}{`numeric(1)`\cr
-#' Fraction parameter of the successive halving algorithm: With every stage the
-#' configuration budget is increased by a factor of `eta` and only the best
-#' `1/eta` points are used for the next stage. Non-integer values are supported,
-#' but `eta` is not allowed to be less or equal 1.
+#' With every stage, the budget is increased by a factor of `eta`
+#' and only the best `1 / eta` points are promoted to the next stage.
+#' Non-integer values are supported, but `eta` is not allowed to be less or
+#' equal 1.
 #' }
 #' \item{`sampler`}{[paradox::Sampler]\cr
-#' Object defining how the samples of the parameter space should be drawn during
-#' the initialization of each bracket. The default is uniform sampling.
+#' Object defining how the samples of the parameter space should be drawn in the
+#' base stage of each bracket. The default is uniform sampling.
 #' }
-#' \item{`repeats`}{`logical(1)`\cr
-#' If `FALSE` (default), hyperband terminates once all brackets are evaluated.
-#' Otherwise, hyperband starts over again once the last bracket is evaluated.
+#' \item{`repetitions`}{`integer(1)`\cr
+#' If `1` (default), optimization is stopped once all brackets are evaluated.
+#' Otherwise, optimization is stopped after `repetitions` runs of hyperband. The
+#' [bbotk::Terminator] might stop the optimization before all repetitions are
+#' executed.
 #' }}
 #'
 #' @section Archive:
@@ -51,50 +62,18 @@
 #'     The bracket index. Counts down to 0.
 #'   * `stage` (`integer(1))`\cr
 #'     The stages of each bracket. Starts counting at 0.
+#'   * `repetition` (`integer(1))`\cr
+#'     Repetition index. Start counting at 1.
 #'
 #' @template section_custom_sampler
-#' @template section_runtime
 #' @template section_progress_bars
-#' @template section_parallelization
 #' @template section_logging
 #
 #' @source
 #' `r format_bib("li_2018")`
 #'
 #' @export
-#' @examples
-#' library(bbotk)
-#' library(data.table)
-#'
-#' search_space = domain = ps(
-#'   x1 = p_dbl(-5, 10),
-#'   x2 = p_dbl(0, 15),
-#'   fidelity = p_dbl(1e-2, 1, tags = "budget")
-#' )
-#'
-#' # modified branin function
-#' objective = ObjectiveRFun$new(
-#'   fun = branin,
-#'   domain = domain,
-#'   codomain = ps(y = p_dbl(tags = "minimize"))
-#' )
-#'
-#' instance = OptimInstanceSingleCrit$new(
-#'   objective = objective,
-#'   search_space = search_space,
-#'   terminator = trm("none")
-#' )
-#'
-#' optimizer = opt("hyperband")
-#'
-#' # modifies the instance by reference
-#' optimizer$optimize(instance)
-#'
-#' # best scoring evaluation
-#' instance$result
-#'
-#' # all evaluations
-#' as.data.table(instance$archive)
+#' @template example_optimizer
 OptimizerHyperband = R6Class("OptimizerHyperband",
   inherit = Optimizer,
   public = list(
@@ -105,9 +84,9 @@ OptimizerHyperband = R6Class("OptimizerHyperband",
       param_set = ps(
         eta     = p_dbl(lower = 1.0001, tags = "required", default = 2),
         sampler = p_uty(custom_check = function(x) check_r6(x, "Sampler", null.ok = TRUE)),
-        repeats = p_lgl(default = FALSE)
+        repetitions = p_int(lower = 1L, default = 1, special_vals = list(Inf))
       )
-      param_set$values = list(eta = 2, sampler = NULL, repeats = FALSE)
+      param_set$values = list(eta = 2, sampler = NULL, repetitions = 1)
 
       super$initialize(
         param_classes = c("ParamLgl", "ParamInt", "ParamDbl", "ParamFct"),
@@ -140,7 +119,7 @@ OptimizerHyperband = R6Class("OptimizerHyperband",
         assert_set_equal(sampler$param_set$ids(), search_space_sampler$ids())
       }
 
-      # r_min is the budget of a single configuration in the first stage
+      # r_min is the budget of a single configuration in the base stage
       # r_max is the maximum budget of a single configuration in the last stage
       # the internal budget is rescaled to a minimum budget of 1
       # for this, the budget is divided by r_min
@@ -162,7 +141,8 @@ OptimizerHyperband = R6Class("OptimizerHyperband",
       # number of configurations in first stages
       n = ceiling((budget / r) * (eta^(0:s_max)) / ((0:s_max) + 1))
 
-      repeat({
+      repetition = 1
+      while (repetition <= pars$repetitions) {
         # original hyperband algorithm iterates over brackets
         # this implementation iterates over stages with same budget
         # the number of iterations (s_max + 1) remains the same in both implementations
@@ -172,8 +152,9 @@ OptimizerHyperband = R6Class("OptimizerHyperband",
           # sample initial configurations of bracket
           xdt = sampler$sample(n[s + 1])$data
           set(xdt, j = budget_id, value = rs)
-          set(xdt, j = "bracket", value = s)
           set(xdt, j = "stage", value = 0)
+          set(xdt, j = "bracket", value = s)
+          set(xdt, j = "repetition", value = repetition)
 
           # promote configurations of previous batch
           if (s != s_max) {
@@ -201,7 +182,7 @@ OptimizerHyperband = R6Class("OptimizerHyperband",
             })
 
             # increase budget and stage
-            xdt_promoted = xdt_promoted[, c(inst$archive$cols_x, "stage", "bracket"), with = FALSE]
+            xdt_promoted = xdt_promoted[, c(inst$archive$cols_x, "stage", "bracket", "repetition"), with = FALSE]
             set(xdt_promoted, j = budget_id, value = rs)
             set(xdt_promoted, j = "stage", value = xdt_promoted[["stage"]] + 1)
 
@@ -211,8 +192,8 @@ OptimizerHyperband = R6Class("OptimizerHyperband",
           if (search_space$class[[budget_id]] == "ParamInt") set(xdt, j = budget_id, value = round(xdt[[budget_id]]))
           inst$eval_batch(xdt)
         }
-        if (!pars$repeats) break
-      })
+        repetition = repetition + 1
+      }
     }
   )
 )
